@@ -4,8 +4,8 @@ use libusb::Context;
 use thiserror::Error;
 
 use crate::device::cfg::{
-    AwgType, Coupling, DeviceFunction, HantekConfig, Probe, Scale, TimeScale, TriggerMode,
-    TriggerSlope,
+    Adjustment, AwgType, Coupling, DeviceFunction, HantekConfig, Probe, RunningStatus, Scale,
+    TimeScale, TrapDuty, TriggerMode, TriggerSlope,
 };
 use crate::device::cmd::{HantekCommandBuilder, RawCommand};
 use crate::device::usb::{HantekUsbDevice, HantekUsbError};
@@ -20,7 +20,7 @@ const READ_ENDPOINT: u8 = 0x80 | 1;
 
 #[derive(Error, Debug)]
 pub enum Hantek2D42Error {
-    #[error("error with usb device")]
+    #[error("error with usb device: {error}")]
     HantekUsbError {
         error: HantekUsbError,
         failed_action: &'static str,
@@ -36,21 +36,38 @@ pub enum Hantek2D42Error {
     TriggerLevelAdjustmentError,
 }
 
+impl Hantek2D42Error {
+    // Because CLion doesn't like the Display implemented by thiserror.
+    pub fn my_to_string(&self) -> impl std::fmt::Display + '_ {
+        self
+    }
+}
+
 pub struct Hantek2D42<'a> {
     pub usb: HantekUsbDevice<'a>,
     config: HantekConfig,
 }
 
 impl<'a> Hantek2D42<'a> {
+    pub fn new(usb: HantekUsbDevice<'a>, config: HantekConfig) -> Self {
+        Self { usb, config }
+    }
+
     pub fn open(context: &'a Context, timeout: Duration) -> Result<Self, Hantek2D42Error> {
-        Ok(Self {
-            usb: HantekUsbDevice::open(context, timeout, (VENDOR_ID__2D42, PRODUCT_ID__2D42))
-                .map_err(|error| Hantek2D42Error::HantekUsbError {
-                    error,
-                    failed_action: "device open",
-                })?,
-            config: HantekConfig::new(timeout, NUM_CHANNELS),
-        })
+        let usb = HantekUsbDevice::open(context, timeout, (VENDOR_ID__2D42, PRODUCT_ID__2D42))
+            .map_err(|error| Hantek2D42Error::HantekUsbError {
+                error,
+                failed_action: "device open",
+            })?;
+        let mut config = HantekConfig::new(NUM_CHANNELS);
+        config.timeout = Some(timeout);
+        Ok(Self::new(usb, config))
+    }
+
+    /// ================================================================= DEVICE
+
+    pub fn get_config(&self) -> &HantekConfig {
+        &self.config
     }
 
     pub fn start(&mut self) -> Result<(), Hantek2D42Error> {
@@ -65,7 +82,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "sending Start command to device",
             })
-            .map(|_| self.config.start())
+            .map(|_| {
+                self.config.running_status = Some(RunningStatus::Start);
+            })
     }
 
     pub fn stop(&mut self) -> Result<(), Hantek2D42Error> {
@@ -80,7 +99,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "sending Stop command to device",
             })
-            .map(|_| self.config.stop())
+            .map(|_| {
+                self.config.running_status = Some(RunningStatus::Stop);
+            })
     }
 
     pub fn set_device_function(&mut self, function: DeviceFunction) -> Result<(), Hantek2D42Error> {
@@ -99,7 +120,7 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting device function",
             })
-            .map(|_| self.config.set_device_function(function))
+            .map(|_| self.config.device_function = Some(function))
     }
 
     /// ================================================================ CHANNEL
@@ -122,7 +143,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "enabling channel",
             })
-            .map(|_| self.config.enable_channel(channel_no))
+            .map(|_| {
+                self.config.enabled_channels.insert(channel_no, Some(true));
+            })
     }
 
     pub fn disable_channel(&mut self, channel_no: usize) -> Result<(), Hantek2D42Error> {
@@ -143,7 +166,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "disabling channel",
             })
-            .map(|_| self.config.disable_channel(channel_no))
+            .map(|_| {
+                self.config.enabled_channels.insert(channel_no, Some(false));
+            })
     }
 
     pub fn set_channel_coupling(
@@ -172,7 +197,11 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting channel coupling",
             })
-            .map(|_| self.config.set_channel_coupling(channel_no, coupling))
+            .map(|_| {
+                self.config
+                    .channel_coupling
+                    .insert(channel_no, Some(coupling));
+            })
     }
 
     pub fn set_channel_probe(
@@ -202,7 +231,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting chanel probe",
             })
-            .map(|_| self.config.set_channel_probe(channel_no, probe))
+            .map(|_| {
+                self.config.channel_probe.insert(channel_no, Some(probe));
+            })
     }
 
     pub fn set_channel_scale(
@@ -239,12 +270,14 @@ impl<'a> Hantek2D42<'a> {
                 failed_action: "setting channel scale",
             })
             .map(|_| {
-                self.config.set_channel_adjustment(
+                self.config.channel_offset_adjustment.insert(
                     channel_no,
-                    4.0 * scale.raw_value(),
-                    -4.0 * scale.raw_value(),
+                    Some(Adjustment::new(
+                        4.0 * scale.raw_value(),
+                        -4.0 * scale.raw_value(),
+                    )),
                 );
-                self.config.set_channel_scale(channel_no, scale);
+                self.config.channel_scale.insert(channel_no, Some(scale));
             })
     }
 
@@ -261,7 +294,7 @@ impl<'a> Hantek2D42<'a> {
         }
         // TODO sanitize offset value range.
 
-        let adjustment = self.config.get_channel_adjustment(channel_no);
+        let adjustment = self.config.channel_offset_adjustment[&channel_no].as_ref();
         if adjustment.is_none() {
             return Err(Hantek2D42Error::ChannelAdjustmentError);
         }
@@ -303,7 +336,11 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting channel offset",
             })
-            .map(|_| self.config.set_channel_offset(channel_no, offset as f32))
+            .map(|_| {
+                self.config
+                    .channel_offset
+                    .insert(channel_no, Some(offset as f32));
+            })
     }
 
     pub fn channel_enable_bandwidth_limit(
@@ -327,7 +364,11 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "enabling channel bandwidth limit",
             })
-            .map(|_| self.config.channel_enable_bandwidth_limit(channel_no))
+            .map(|_| {
+                self.config
+                    .channel_bandwidth_limit
+                    .insert(channel_no, Some(true));
+            })
     }
 
     pub fn channel_disable_bandwidth_limit(
@@ -351,7 +392,11 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "disabling channel bandwidth limit",
             })
-            .map(|_| self.config.channel_disable_bandwidth_limit(channel_no))
+            .map(|_| {
+                self.config
+                    .channel_bandwidth_limit
+                    .insert(channel_no, Some(false));
+            })
     }
 
     pub fn capture(
@@ -456,9 +501,9 @@ impl<'a> Hantek2D42<'a> {
                 failed_action: "setting time scale",
             })
             .map(|_| {
-                self.config
-                    .set_time_offset_adjustment(15.0 * (raw as f32), -15.0 * (raw as f32));
-                self.config.set_time_scale(time_scale);
+                self.config.time_offset_adjustment =
+                    Some(Adjustment::new(15.0 * (raw as f32), -15.0 * (raw as f32)));
+                self.config.time_scale = Some(time_scale);
             })
     }
 
@@ -470,7 +515,7 @@ impl<'a> Hantek2D42<'a> {
             panic!("bad value for time_offset={}", time_offset);
         }
 
-        let adjustment = self.config.get_time_offset_adjustment();
+        let adjustment = self.config.time_offset_adjustment.as_ref();
         if adjustment.is_none() {
             return Err(Hantek2D42Error::TimeOffsetAdjustmentError);
         }
@@ -503,15 +548,16 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting time offset",
             })
-            .map(|_| self.config.set_time_offset(time_offset as f32))
+            .map(|_| {
+                self.config.time_offset = Some(time_offset as f32);
+            })
     }
 
     pub fn set_trigger_source(&mut self, channel_no: usize) -> Result<(), Hantek2D42Error> {
         self.assert_channel_no(channel_no);
 
-        let scale = self
-            .config
-            .get_channel_scale(channel_no)
+        let scale = self.config.channel_scale[&channel_no]
+            .as_ref()
             .map(|it| it.raw_value());
         if scale.is_none() {
             return Err(Hantek2D42Error::TriggerLevelAdjustmentError);
@@ -530,9 +576,9 @@ impl<'a> Hantek2D42<'a> {
                 failed_action: "setting trigger source",
             })
             .map(|_| {
-                self.config.set_trigger_source_channel_no(channel_no);
-                self.config
-                    .set_trigger_level_adjustment(4.0 * scale, -4.0 * scale);
+                self.config.trigger_source_channel = Some(channel_no);
+                self.config.trigger_level_adjustment =
+                    Some(Adjustment::new(4.0 * scale, -4.0 * scale));
             })
     }
 
@@ -555,7 +601,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting trigger slope",
             })
-            .map(|_| self.config.set_trigger_slope(trigger_slope))
+            .map(|_| {
+                self.config.trigger_slope = Some(trigger_slope);
+            })
     }
 
     pub fn set_trigger_mode(&mut self, trigger_mode: TriggerMode) -> Result<(), Hantek2D42Error> {
@@ -574,7 +622,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting trigger mode",
             })
-            .map(|_| self.config.set_trigger_mode(trigger_mode))
+            .map(|_| {
+                self.config.trigger_mode = Some(trigger_mode);
+            })
     }
 
     pub fn set_trigger_level_with_auto_adjustment(
@@ -588,7 +638,7 @@ impl<'a> Hantek2D42<'a> {
             );
         }
 
-        let adjustment = self.config.get_trigger_level_adjustment();
+        let adjustment = self.config.trigger_level_adjustment.as_ref();
         if adjustment.is_none() {
             return Err(Hantek2D42Error::TriggerLevelAdjustmentError);
         }
@@ -619,7 +669,7 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting trigger level",
             })
-            .map(|_| self.config.set_trigger_level(trigger_level as f32))
+            .map(|_| self.config.trigger_level = Some(trigger_level as f32))
     }
 
     ///=================================================================== AWG
@@ -645,7 +695,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting awg mode",
             })
-            .map(|_| self.config.set_awg_type(awg_type))
+            .map(|_| {
+                self.config.awg_type = Some(awg_type);
+            })
     }
 
     pub fn set_awg_frequency(&mut self, frequency: f32) -> Result<(), Hantek2D42Error> {
@@ -662,7 +714,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting awg frequency",
             })
-            .map(|_| self.config.set_awg_frequency(frequency))
+            .map(|_| {
+                self.config.awg_frequency = Some(frequency);
+            })
     }
 
     pub fn set_awg_amplitude(&mut self, amplitude: f32) -> Result<(), Hantek2D42Error> {
@@ -685,7 +739,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting awg amplitude",
             })
-            .map(|_| self.config.set_awg_amplitude(amplitude))
+            .map(|_| {
+                self.config.awg_amplitude = Some(amplitude);
+            })
     }
 
     pub fn set_awg_offset(&mut self, offset: f32) -> Result<(), Hantek2D42Error> {
@@ -708,7 +764,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting awg offset",
             })
-            .map(|_| self.config.set_awg_offset(offset))
+            .map(|_| {
+                self.config.awg_offset = Some(offset);
+            })
     }
 
     pub fn set_awg_duty_square(&mut self, duty: f32) -> Result<(), Hantek2D42Error> {
@@ -726,7 +784,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting awg square duty",
             })
-            .map(|_| self.config.set_awg_duty_square(duty))
+            .map(|_| {
+                self.config.awg_duty_square = Some(duty);
+            })
     }
 
     pub fn set_awg_duty_ramp(&mut self, duty: f32) -> Result<(), Hantek2D42Error> {
@@ -745,7 +805,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting awg ramp duty",
             })
-            .map(|_| self.config.set_awg_duty_ramp(duty))
+            .map(|_| {
+                self.config.awg_duty_ramp = Some(duty);
+            })
     }
 
     pub fn set_awg_duty_trap(
@@ -771,7 +833,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "setting awg ramp duty",
             })
-            .map(|_| self.config.set_awg_duty_trap(high, low, rise))
+            .map(|_| {
+                self.config.awg_duty_trap = Some(TrapDuty { high, low, rise });
+            })
     }
 
     pub fn awg_start(&mut self) -> Result<(), Hantek2D42Error> {
@@ -786,7 +850,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "awg start",
             })
-            .map(|_| self.config.awg_start())
+            .map(|_| {
+                self.config.awg_running_status = Some(RunningStatus::Start);
+            })
     }
 
     pub fn awg_stop(&mut self) -> Result<(), Hantek2D42Error> {
@@ -801,7 +867,9 @@ impl<'a> Hantek2D42<'a> {
                 error,
                 failed_action: "awg stop",
             })
-            .map(|_| self.config.awg_stop())
+            .map(|_| {
+                self.config.awg_running_status = Some(RunningStatus::Stop);
+            })
     }
 
     ///=============================================================== INTERNAL
